@@ -1,12 +1,12 @@
 package Minion::Backend::Mango;
 use Mojo::Base 'Minion::Backend';
 
-our $VERSION = '0.99_50';
+our $VERSION = '0.90';
 
 use Mango;
-use Mango::BSON qw(bson_oid bson_time bson_doc);
+use Mango::BSON qw(bson_oid bson_time);
+use Scalar::Util 'weaken';
 use Sys::Hostname 'hostname';
-use Time::HiRes 'time';
 
 has 'mango';
 has jobs          => sub { $_[0]->mango->db->collection($_[0]->prefix . '.jobs') };
@@ -14,18 +14,20 @@ has notifications => sub { $_[0]->mango->db->collection($_[0]->prefix . '.notifi
 has prefix        => 'minion';
 has workers       => sub { $_[0]->mango->db->collection($_[0]->prefix . '.workers') };
 
-sub dequeue {
-  my ($self, $wid, $wait, $options) = @_;
 
+sub dequeue {
+  my ($self, $oid, $timeout) = @_;
+
+  # Capped collection for notifications
   $self->_notifications;
 
-  my $end = time + $wait;
-  do {
-    if (my $job = $self->_try($wid, $options)) { return $self->_job_info($job); }
-    $self->_await;
-  } while time < $end;
+  # Await notifications
+  my $end = bson_time->to_epoch + $timeout;
+  my $job;
+  do { $self->_await and $job = $self->_try($oid) } while !$job && bson_time->to_epoch < $end;
 
-  return $self->_job_info($self->_try($wid, $options));
+  return undef unless $self->_job_info($job ||= $self->_try($oid));
+  return {args => $job->{args}, id => $job->{_id}, task => $job->{task}};
 }
 
 sub enqueue {
@@ -33,20 +35,19 @@ sub enqueue {
   my $args    = shift // [];
   my $options = shift // {};
 
+  # Capped collection for notifications
   $self->_notifications;
 
   my $doc = {
-    args     => $args,
-    attempts => $options->{attempts} // 1,
-    created  => bson_time,
-    delayed  => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
+    args    => $args,
+    created => bson_time,
+    delayed => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
     priority => $options->{priority} // 0,
-    queue    => $options->{queue}    // 'default',
-    retries  => 0,
     state    => 'inactive',
     task     => $task
   };
 
+  # Blocking
   my $oid = $self->jobs->insert($doc);
   $self->notifications->insert({c => 'created'});
   return $oid;
@@ -56,9 +57,7 @@ sub fail_job { shift->_update(1, @_) }
 
 sub finish_job { shift->_update(0, @_) }
 
-sub job_info {
-  $_[0]->_job_info($_[0]->jobs->find_one(bson_oid($_[1])));
-}
+sub job_info { $_[0]->_job_info($_[0]->jobs->find_one(bson_oid($_[1]))) }
 
 sub list_jobs {
   my ($self, $skip, $limit, $options) = @_;
@@ -81,15 +80,7 @@ sub list_workers {
 sub new { shift->SUPER::new(mango => Mango->new(@_)) }
 
 sub register_worker {
-  my ($self, $id) = @_;
-
-  return $id
-    if $id
-    && $self->workers->find_and_modify(
-    {query => {_id => $id}, update => {'$set' => {notified => bson_time}}});
-
-  $self->jobs->ensure_index(bson_doc(state => 1, delayed => 1, task => 1, queue => 1));
-  $self->workers->insert({host => hostname, pid => $$, started => bson_time, notified => bson_time});
+  shift->workers->insert({host => hostname, pid => $$, started => bson_time});
 }
 
 sub remove_job {
@@ -99,42 +90,38 @@ sub remove_job {
 }
 
 sub repair {
-  my $self   = shift;
-  my $minion = $self->minion;
+  my $self = shift;
 
-  # Check worker registry
+  # Check workers on this host (all should be owned by the same user)
   my $workers = $self->workers;
-  $workers->remove({notified => {'$lt' => bson_time((time - $minion->missing_after) * 1000)}});
+  my $cursor = $workers->find({host => hostname});
+  while (my $worker = $cursor->next) {
+    $workers->remove($worker->{_id}) unless kill 0, $worker->{pid};
+  }
 
   # Abandoned jobs
   my $jobs = $self->jobs;
-  my $cursor = $jobs->find({state => 'active'}, {_id => 1, retries => 1, worker => 1});
+  $cursor = $jobs->find({state => 'active'});
   while (my $job = $cursor->next) {
-    $self->fail_job($job->{_id}, $job->{retries}, 'Worker went away')
+    $jobs->save({%$job, state => 'failed', error => 'Worker went away'})
       unless $workers->find_one($job->{worker});
   }
 
   # Old jobs
-  $jobs->remove(
-    {state => 'finished', finished => {'$lt' => bson_time((time - $minion->remove_after) * 1000)}});
+  my $after = bson_time((time - $self->minion->remove_after) * 1000);
+  $jobs->remove({state => 'finished', finished => {'$lt' => $after}});
 }
 
 sub reset { $_->options && $_->drop for $_[0]->workers, $_[0]->jobs }
 
 sub retry_job {
-  my ($self, $oid, $retries) = (shift, shift, shift);
-  my $options = shift // {};
+  my ($self, $oid) = @_;
 
-  my $query = {_id => $oid, retries => $retries, state => {'$in' => [qw(failed finished inactive)]}};
+  my $query = {_id => $oid, state => {'$in' => [qw(failed finished)]}};
   my $update = {
     '$inc' => {retries => 1},
-    '$set' => {
-      retried => bson_time,
-      state   => 'inactive',
-      delayed => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
-      (defined $options->{priority} ? (priority => $options->{priority}) : ()),
-      (defined $options->{queue}    ? (queue    => $options->{queue})    : ())
-    }
+    '$set' => {retried => bson_time, state => 'inactive'},
+    '$unset' => {map { $_ => '' } qw(error finished result started worker)}
   };
 
   return !!$self->jobs->update($query, $update)->{n};
@@ -159,9 +146,9 @@ sub worker_info { $_[0]->_worker_info($_[0]->workers->find_one($_[1])) }
 sub _await {
   my $self = shift;
 
-  my $last = $self->{last} //= bson_oid;
-  my $cursor
-    = $self->notifications->find({_id => {'$gt' => $last}, c => 'created'})->tailable(1)->await_data(1);
+  my $last = $self->{last} //= bson_oid(0 x 24);
+  my $cursor =
+    $self->notifications->find({_id => {'$gt' => $last}, c => 'created'})->tailable(1)->await_data(1);
   return undef unless my $doc = $cursor->next || $cursor->next;
   $self->{last} = $doc->{_id};
   return 1;
@@ -173,14 +160,12 @@ sub _job_info {
   return undef unless my $job = shift;
   return {
     args     => $job->{args},
-    attempts => $job->{attempts},
     created  => $job->{created} ? $job->{created}->to_epoch : undef,
     delayed  => $job->{delayed} ? $job->{delayed}->to_epoch : undef,
+    error    => $job->{error},
     finished => $job->{finished} ? $job->{finished}->to_epoch : undef,
     id       => $job->{_id},
     priority => $job->{priority},
-    queue    => $job->{queue},
-    result   => $job->{result},
     retried  => $job->{retried} ? $job->{retried}->to_epoch : undef,
     retries => $job->{retries} // 0,
     started => $job->{started} ? $job->{started}->to_epoch : undef,
@@ -196,23 +181,22 @@ sub _notifications {
   $self->{capped} ? return : $self->{capped}++;
   my $notifications = $self->notifications;
   return if $notifications->options;
-  $notifications->create({capped => \1, size => 1048576, max => 128});
+  $notifications->create({capped => \1, max => 8, size => 1048576});
   $notifications->insert({});
 }
 
 sub _try {
-  my ($self, $wid, $options) = @_;
+  my ($self, $oid) = @_;
 
   my $doc = {
-    query => bson_doc(
-      state   => 'inactive',
+    query => {
       delayed => {'$lt' => bson_time},
-      task    => {'$in' => [keys %{$self->minion->tasks}]},
-      queue   => {'$in' => $options->{queues} || ['default']}
-    ),
-    fields => bson_doc(args     => 1,  retries => 1, task => 1),
-    sort   => bson_doc(priority => -1, created => 1),
-    update => {'$set' => {started => bson_time, state => 'active', worker => $wid}},
+      state   => 'inactive',
+      task    => {'$in' => [keys %{$self->minion->tasks}]}
+    },
+    fields => {args     => 1, task => 1},
+    sort   => {priority => -1},
+    update => {'$set' => {started => bson_time, state => 'active', worker => $oid}},
     new    => 1
   };
 
@@ -220,17 +204,12 @@ sub _try {
 }
 
 sub _update {
-  my ($self, $fail, $oid, $retries, $result) = @_;
+  my ($self, $fail, $oid, $err) = @_;
 
-  my $update = {finished => bson_time, result => $result, state => $fail ? 'failed' : 'finished'};
-  my $query = {_id => $oid, retries => $retries, state => 'active'};
-
-  my $opts = {query => $query, update => {'$set' => $update}, fields => {attempts => 1}};
-  return undef unless my $job = $self->jobs->find_and_modify($opts);
-  return 1 if !$fail || (my $attempts = $job->{attempts}) == 1;
-  return 1 if $retries >= ($attempts - 1);
-  my $delay = $self->minion->backoff->($retries);
-  return $self->retry_job($oid, $retries, {delay => $delay});
+  my $update = {finished => bson_time, state => $fail ? 'failed' : 'finished'};
+  $update->{error} = $err if $fail;
+  my $query = {_id => $oid, state => 'active'};
+  return !!$self->jobs->update($query, {'$set' => $update})->{n};
 }
 
 sub _worker_info {
@@ -239,12 +218,11 @@ sub _worker_info {
   return undef unless my $worker = shift;
   my $cursor = $self->jobs->find({state => 'active', worker => $worker->{_id}});
   return {
-    host     => $worker->{host},
-    id       => $worker->{_id},
-    jobs     => [map { $_->{_id} } @{$cursor->all}],
-    pid      => $worker->{pid},
-    started  => $worker->{started}->to_epoch,
-    notified => $worker->{notified}->to_epoch
+    host    => $worker->{host},
+    id      => $worker->{_id},
+    jobs    => [map { $_->{_id} } @{$cursor->all}],
+    pid     => $worker->{pid},
+    started => $worker->{started}->to_epoch
   };
 }
 
@@ -312,24 +290,11 @@ L<Minion::Backend::Mango> inherits all methods from L<Minion::Backend> and imple
 
 =head2 dequeue
 
-  my $job_info = $backend->dequeue($worker_id, 0.5);
-  my $job_info = $backend->dequeue($worker_id, 0.5, {queues => ['default']});
+  my $info = $backend->dequeue($worker_id, 0.5);
 
 Wait for job, dequeue it and transition from C<inactive> to C<active> state or
 return C<undef> if queue was empty.
 
-These options are currently available:
-
-=over 2
-
-=item queues
-
-  queues => ['important']
-
-One or more queues to dequeue jobs from, defaults to C<default>.
-
-=back
-
 =head2 enqueue
 
   my $job_id = $backend->enqueue('foo');
@@ -340,12 +305,6 @@ Enqueue a new job with C<inactive> state. These options are currently available:
 
 =over 2
 
-=item attempts
-
-  attempts => 25
-
-Number of times performing this job will be attempted, defaults to C<1>.
-
 =item delay
 
   delay => 10
@@ -358,12 +317,6 @@ Delay job for this many seconds from now.
 
 Job priority, defaults to C<0>.
 
-=item queue
-
-  queue => 'important'
-
-Queue to put job in, defaults to C<default>.
-
 =back
 
 =head2 fail_job
@@ -425,9 +378,8 @@ Construct a new L<Minion::Backend::Mango> object.
 =head2 register_worker
 
   my $worker_id = $backend->register_worker;
-  my $worker_id = $backend->register_worker($worker_id);
 
-Register worker or send heartbeat to show that this worker is still alive.
+Register worker.
 
 =head2 remove_job
 
@@ -450,34 +402,8 @@ Reset job queue.
 =head2 retry_job
 
   my $bool = $backend->retry_job($job_id);
-  my $bool = $backend->retry_job($job_id, {delay => 10});
 
-Transition from C<failed> or C<finished> state back to C<inactive>, already
-C<inactive> jobs may also be retried to change options.
-
-These options are currently available:
-
-=over 2
-
-=item delay
-
-  delay => 10
-
-Delay job for this many seconds (from now).
-
-=item priority
-
-  priority => 5
-
-Job priority.
-
-=item queue
-
-  queue => 'important'
-
-Queue to put job in.
-
-=back
+Transition from C<failed> or C<finished> state back to C<inactive>.
 
 =head2 stats
 
@@ -507,7 +433,7 @@ Sebastian Riedel E<lt>sri@cpan.orgE<gt>
 
 Copyright (C) 2014, Sebastian Riedel.
 
-Copyright (C) 2015-2016, Andrey Khozov.
+Copyright (C) 2015, Andrey Khozov.
 
 This library is free software; you can redistribute it and/or modify it under the same terms as Perl itself.
 
